@@ -5,11 +5,19 @@ Rewrite Craftr DSL code to pure Python code.
 import contextlib
 import enum
 import logging
+import re
 import string
 import sys
 import typing as t
 from dataclasses import dataclass
-from nr.util.parsing import rules, Cursor, ProxyToken as _ProxyToken, RuleSet, Tokenizer
+
+from nr.util.parsing import (
+  Cursor,
+  ProxyToken as _ProxyToken,
+  RuleSet,
+  Tokenizer,
+  rules,
+)
 
 try:
   from termcolor import colored
@@ -18,6 +26,7 @@ except ImportError:
   def colored(s, *a, **kw) -> str:  # type: ignore
     return str(s)
 
+from ._util import debug_trace
 
 logger = logging.getLogger(__name__)
 
@@ -41,6 +50,12 @@ class Grammar:
   #: syntax feature when enabling the #NameRewriter with #TranspileOptions.closure_target.
   local_def: bool = False
 
+  #: The keyword for local definitions.
+  local_keyword: str = 'def'
+
+  #: The prefix for variable names that are localized. This is picked up by the transpiler.
+  local_prefix: str = '_def_'
+
 
 class Token(enum.Enum):
   Eof = enum.auto()
@@ -53,19 +68,29 @@ class Token(enum.Enum):
   Control = enum.auto()
 
 
+PYTHON_BLOCK_KEYWORDS = frozenset(['class', 'def', 'if', 'elif', 'else', 'for', 'while', 'with'])
+ASSIGNMENT_OPERATORS = ['=', '+=', '-=', '*=', '/=', '%=', '//=', '**=', '&=', '|=', '^=', '>>=', '<<=', '@=']
+BINARY_OPERATORS = [x[:-1] for x in ASSIGNMENT_OPERATORS[1:]] + [
+  '.', '<', '>', '==', '<=', '>=', ':=', 'is', 'and', 'or']
+UNARY_OPERATORS = ['not', '~']
+OTHER_CONTROL_CHARACTERS = list('()[]{},:;') + ['->']
+_ALL_CONTROL_CHARACTERS = sorted(
+  ASSIGNMENT_OPERATORS + BINARY_OPERATORS + UNARY_OPERATORS + OTHER_CONTROL_CHARACTERS,
+  key=len,
+  reverse=True,
+)
+_WORD_CONTROL_CHARACTERS = [op for op in _ALL_CONTROL_CHARACTERS if op.isalpha()]
+
 rule_set = RuleSet((Token.Eof, ''))
 rule_set.rule(Token.Indent, rules.regex_extract(r'[\t ]*', at_line_start_only=True))
 rule_set.rule(Token.Newline, rules.regex_extract(r'\n'))
 rule_set.rule(Token.Whitespace, rules.regex_extract(r'\s+'))
 rule_set.rule(Token.Comment, rules.regex_extract(r'#.*'))
-rule_set.rule(Token.Control, rules.regex_extract(r'is\b|not\b'))
+rule_set.rule(Token.Control, rules.regex_extract('(' + '|'.join(map(re.escape, _WORD_CONTROL_CHARACTERS)) + r')\b'))
 rule_set.rule(Token.Name, rules.regex_extract(r'[A-Za-z\_][A-Za-z0-9\_]*'))
 rule_set.rule(Token.Literal, rules.regex_extract(r'[+\-]?(\d+)(\.\d*)?'))
 rule_set.rule(Token.Literal, rules.string_literal())
-rule_set.rule(
-    Token.Control,
-    rules.regex_extract(
-        r'(\[|\]|\{|\}|\(|\)|<<|<|>>|>|\.|,|\->|\-|!|\+|\*\*|\*|//|/|->|==|<=|>=|<|>|=|:|&|\||^|%|@|;)'))
+rule_set.rule(Token.Control, rules.regex_extract('|'.join(map(re.escape, _ALL_CONTROL_CHARACTERS))))
 
 
 class ProxyToken(_ProxyToken):
@@ -119,7 +144,7 @@ class SyntaxError(Exception):
     lines = [
         '',
         f'  in {colored(self.filename, "blue")} at line {self.line}: {colored(self.message, "red")}',
-        *('  |' + l for l in self.get_text_hint().splitlines()),
+        *('  |' + line for line in self.get_text_hint().splitlines()),
     ]
     return '\n'.join(lines)
 
@@ -172,11 +197,6 @@ class Rewriter:
   tact as much as possible (not always fully accurate).
   """
 
-  UNARY_OPERATORS = frozenset(['not', '~'])
-  BINARY_OPERATORS = frozenset(
-      ['-', '+', '*', '**', '/', '//', '^', '|', '&', '.', '==', '<=', '>=', '<', '>', 'is', '%'])
-  PYTHON_BLOCK_KEYWORDS = frozenset(['class', 'def', 'if', 'elif', 'else', 'for', 'while', 'with'])
-
   def __init__(self, text: str, filename: str, grammar: t.Optional[Grammar] = None) -> None:
     """
     # Arguments
@@ -221,6 +241,7 @@ class Rewriter:
     text = self.tokenizer.scanner.getline(pos)
     return SyntaxError(msg, self.filename, pos.line, pos.column, text)
 
+  @debug_trace
   def _consume_whitespace(self, newlines: t.Union[bool, ParseMode] = False, reset_to_indent: bool = True) -> str:
     """
     Consumes whitespace, indents, comments, and, if enabled, newlines until a different token is
@@ -243,6 +264,7 @@ class Rewriter:
       parts.pop()
     return ''.join(parts)
 
+  @debug_trace
   def _parse_closure(self) -> t.Optional[Closure]:
     """
     Attempts to parse a closure at the current position of the tokenizer. Closures can have the
@@ -294,6 +316,7 @@ class Rewriter:
     self._closure_counter += 1
     return Closure(closure_id, pos.line, pos.column, arglist, body, expr)
 
+  @debug_trace
   def _parse_closure_body(self) -> t.Optional[str]:
     """
     Parses the body of a closure and returns it's code. Expects the tokenizer to point to the
@@ -317,6 +340,7 @@ class Rewriter:
     token.next()
     return code
 
+  @debug_trace
   def _parse_closure_header(self) -> t.Optional[t.List[str]]:
     """
     Handles the possible formats for a closure header, f.e. a single argument name or an arglist
@@ -345,6 +369,7 @@ class Rewriter:
       token.next()
       return arglist
 
+  @debug_trace
   def _parse_closure_arglist(self) -> t.Optional[t.List[str]]:
     """
     This method expects an open parenthesis as the current token and attempts to extract a list of
@@ -355,7 +380,6 @@ class Rewriter:
     assert token.tv == (Token.Control, '('), token
 
     state = token.save()
-    begin = token.pos.offset
     with token.set_skipped({Token.Whitespace, Token.Comment, Token.Newline, Token.Indent}):
       token.next()
 
@@ -364,8 +388,10 @@ class Rewriter:
 
       while token.tv != (Token.Control, ')'):
 
-        if (not is_delimited  # Token is not preceeded by an opening parentheses or comma.
-            or token.type != Token.Name):  # We can only accept a name at this position.
+        if (
+          not is_delimited  # Token is not preceeded by an opening parentheses or comma.
+          or token.type != Token.Name
+        ):  # We can only accept a name at this position.
           token.load(state)
           return None
 
@@ -380,6 +406,7 @@ class Rewriter:
       token.next()
       return arglist
 
+  @debug_trace
   def _rewrite_expr(self, mode: ParseMode) -> str:
     """
     Consumes a Python expression and returns it's code. Does not parse over a comma.
@@ -395,7 +422,7 @@ class Rewriter:
     while token:
       code += self._consume_whitespace(mode)
 
-      if token.type == Token.Control and token.value in self.BINARY_OPERATORS:
+      if token.type == Token.Control and token.value in BINARY_OPERATORS:
         code += token.value
         token.next()
         code += self._rewrite_expr(mode)
@@ -410,18 +437,46 @@ class Rewriter:
 
     return code
 
+  @debug_trace
+  def _find_current_line_indent(self) -> int:
+    with self._lookahead():
+      self.tokenizer.scanner.pos = self.tokenizer.scanner.pos.get_line_begin()
+      token = self.tokenizer.next()
+      assert token.type == Token.Indent
+      return len(token.value)
+
+  @debug_trace
+  def _check_next_indent(self, min_indent: int) -> t.Union[int, None]:
+    with self._lookahead():
+      token = ProxyToken(self.tokenizer)
+      assert token.type in (Token.Newline, Token.Indent), token
+      whitespace = self._consume_whitespace(True, False).splitlines()
+      indent = len(whitespace[-1])
+      if indent < min_indent:
+        return None
+      return indent
+
+  @debug_trace
   def _rewrite_items(self, mode: ParseMode) -> str:
     """
     Rewrites expressions separated by commas.
     """
 
+    line_indent = self._find_current_line_indent()
+    continuation_indent: t.Union[int, None] = None
+
     token = ProxyToken(self.tokenizer)
     code = ''
     upsert_comma = False
+    upgraded_to_call_args = False
+    do_break = False
+
     while True:
       code += self._consume_whitespace(mode)
-      if token.type == Token.Newline and not mode & ParseMode.GROUPED:
+
+      if token.type == Token.Newline and not (mode & ParseMode.GROUPED):
         break
+
       with self._lookahead() as commit:
         try:
           code += self._rewrite_expr(mode=mode)
@@ -429,27 +484,56 @@ class Rewriter:
           logger.debug('syntax error consumed while trying to parse expression', exc_info=sys.exc_info())
           break
         commit()
+
       code += self._consume_whitespace(mode)
+
       if mode & ParseMode.CALL_ARGS and (token.is_control('=') or
                                          (self.grammar.colon_kwargs and token.is_control(':'))):
         code += '='
         token.next()
         # TODO(NiklasRosenstein): This may be problematic in unparenthesised calls?
         code += self._rewrite_expr(mode=mode)
+
       if token.is_control(','):
         code += ','
         token.next()
         upsert_comma = False
+
       elif mode & ParseMode.CALL_ARGS and self.grammar.nocomma_args:
         code += ','
         upsert_comma = True
+
       else:
+        do_break = True
+
+      if token.type == Token.Newline:
+        if mode & ParseMode.GROUPED:
+          pass
+
+        elif continuation_indent is None:
+          continuation_indent = self._check_next_indent(line_indent + 1)
+          if continuation_indent is None:
+            break
+          if mode == ParseMode.DEFAULT and self.grammar.unparen_calls:
+            mode = ParseMode.CALL_ARGS
+            code += '('
+            upgraded_to_call_args = True
+
+        elif self._check_next_indent(continuation_indent) is None:
+          break
+        code += self._consume_whitespace(True)
+        continue
+
+      if do_break:
         break
 
     if upsert_comma:
       code = code.rstrip(',')
+    if upgraded_to_call_args:
+      code += ')'
     return code
 
+  @debug_trace
   def _rewrite_atom(self, mode: ParseMode = ParseMode.DEFAULT) -> str:
     """
     Consumes a Python or Craftr DSL language atom and returns it rewritten as pure Python code. If
@@ -463,7 +547,8 @@ class Rewriter:
       return self._rewrite_dict()
 
     code = ''
-    if closure := self._parse_closure():
+    closure = self._parse_closure()
+    if closure:
       code += closure.id
       self._closures[closure.id] = closure
 
@@ -493,7 +578,7 @@ class Rewriter:
       code += token.value
       token.next()
 
-    elif token.type == Token.Control and token.value in self.UNARY_OPERATORS:
+    elif token.type == Token.Control and token.value in UNARY_OPERATORS:
       code += token.value
       token.next()
       code += self._rewrite_expr(mode=mode)
@@ -504,6 +589,7 @@ class Rewriter:
 
     return code
 
+  @debug_trace
   def _test_dict(self) -> bool:
     """
     Tests if the code from the current opening curly brace looks like a dictionary definition.
@@ -523,6 +609,7 @@ class Rewriter:
       except SyntaxError:
         return False
 
+  @debug_trace
   def _rewrite_dict(self) -> str:
     token = ProxyToken(self.tokenizer)
     assert token.is_control('{'), token
@@ -552,6 +639,7 @@ class Rewriter:
     token.next()
     return code + '}'
 
+  @debug_trace
   def _rewrite_stmt_singleline(self) -> str:
     token = ProxyToken(self.tokenizer)
     code = self._consume_whitespace(False)
@@ -582,6 +670,7 @@ class Rewriter:
     else:
       return code + self._rewrite_stmt_line_expr_or_assign()
 
+  @debug_trace
   def _rewrite_stmt_line_expr_or_assign(self) -> str:
     token = ProxyToken(self.tokenizer)
     code = self._rewrite_items(ParseMode.DEFAULT)
@@ -596,9 +685,10 @@ class Rewriter:
 
     code += self._consume_whitespace(newlines=False)
 
-    if token.tv == (Token.Control, '='):  # Assignment
+    if token.type == Token.Control and token.value in ASSIGNMENT_OPERATORS:
+      op = token.value
       token.next()
-      code += '=' + self._consume_whitespace(newlines=False) + self._rewrite_items(ParseMode.DEFAULT)
+      code += op + self._consume_whitespace(newlines=False) + self._rewrite_items(ParseMode.DEFAULT)
 
     elif token and not token.is_ignorable(True) and not token.is_control(')]}:') and self.grammar.unparen_calls:
       if code[-1].isspace():
@@ -618,6 +708,7 @@ class Rewriter:
 
     return code + self._consume_whitespace(True)
 
+  @debug_trace
   def _test_local_def(self) -> t.Optional[str]:
     """
     Tests if the current `def` keyword introduces a local variable assignment, and if so,
@@ -625,14 +716,14 @@ class Rewriter:
     """
 
     token = ProxyToken(self.tokenizer)
-    assert token.tv == (Token.Name, 'def'), token
+    assert token.tv == (Token.Name, self.grammar.local_keyword), token
 
     with self._lookahead() as commit:
       token.next()
       self._consume_whitespace(False)
       if token.type != Token.Name:
         return None
-      code = '_def_' + token.value
+      code = self.grammar.local_prefix + token.value
       token.next()
       code += self._consume_whitespace(False)
       if not token.is_control('='):
@@ -643,6 +734,7 @@ class Rewriter:
       commit()
       return code
 
+  @debug_trace
   def _rewrite_stmt(self, indentation: int) -> str:
     """
     Parses a line statement of Python code. Returns an empty string if the actual indendation of
@@ -662,10 +754,12 @@ class Rewriter:
     code += token.value
     token.next()
 
-    if self.grammar.local_def and token.tv == (Token.Name, 'def') and (defcode := self._test_local_def()):
-      return code + defcode
+    if self.grammar.local_def and token.tv == (Token.Name, self.grammar.local_keyword):
+      defcode = self._test_local_def()
+      if defcode:
+        return code + defcode
 
-    if token.type == Token.Name and token.value in self.PYTHON_BLOCK_KEYWORDS:
+    if token.type == Token.Name and token.value in PYTHON_BLOCK_KEYWORDS:
       # Parse to the next colon.
       # TODO(nrosenstein): If we want to support Craftr DSL syntax in the expressions of block
       #   statements, we'll need to rewrite them on a more granular level.
@@ -685,6 +779,7 @@ class Rewriter:
     else:
       return code + self._rewrite_stmt_singleline()
 
+  @debug_trace
   def _rewrite_stmt_block(self, parent_indentation: t.Optional[int] = None) -> str:
     """
     Rewrites an entire statement block and returns it's rewritten code.
@@ -698,10 +793,14 @@ class Rewriter:
     if parent_indentation is not None and len(token.value) <= parent_indentation:
       raise self._syntax_error(f'expected indent > {parent_indentation}, found {token}')
     indentation = len(token.value)
-    while token and (stmt := self._rewrite_stmt(indentation)):
+    while token:
+      stmt = self._rewrite_stmt(indentation)
+      if not stmt:
+        break
       code += stmt + self._consume_whitespace(True)
     return code
 
+  @debug_trace
   def rewrite(self) -> RewriteResult:
     """
     Rewrite the code and return the #RewriteResult. This can be interpreted by the
